@@ -2,6 +2,7 @@
 from __future__ import print_function
 
 import ast
+from compiler.visitor import ASTVisitor
 from compiler.consts import SC_LOCAL, SC_GLOBAL_IMPLICIT, SC_GLOBAL_EXPLICIT, \
     SC_FREE, SC_CELL, SC_UNKNOWN
 from compiler.misc import mangle
@@ -32,6 +33,7 @@ class Scope:
         self.cells = {}
         self.children = []
         self.parent = None
+        self.coroutine = False
         # nested is true if the class could contain free variables,
         # i.e. if it is nested within another function.
         self.nested = None
@@ -39,7 +41,7 @@ class Scope:
         # but explicitly mark it as global. Bytecode-wise, this is handled
         # automagically, but we need to generate proper __qualname__ for these.
         self.global_scope = False
-        self.generator = None
+        self.generator = False
         self.klass = None
         if klass is not None:
             for i in range(len(klass)):
@@ -96,7 +98,7 @@ class Scope:
     def get_children(self):
         return self.children
 
-    def DEBUG(self):
+    def dump(self):
         print(self.name, self.nested and "nested" or "")
         print("\tglobals: ", self.globals)
         print("\texplicit_globals: ", self.explicit_globals)
@@ -105,6 +107,15 @@ class Scope:
         print("\tuses: ", self.uses)
         print("\tfrees:", self.frees)
 
+    # Legacy method name from the original Python2 module,
+    # for backward compatibility.
+    DEBUG = dump
+
+    def dump_recursive(self):
+        self.dump()
+        for s in self.children:
+            s.dump_recursive()
+
     def check_name(self, name):
         """Return scope of name.
 
@@ -120,7 +131,7 @@ class Scope:
             return SC_FREE
         if name in self.defs:
             return SC_LOCAL
-        if self.nested and (name in self.frees or name in self.uses):
+        if self.nested and name in self.uses:
             return SC_FREE
         if self.nested:
             return SC_UNKNOWN
@@ -129,10 +140,13 @@ class Scope:
 
     def get_free_vars(self):
         if not self.nested:
-            assert len(self.frees) <= 1
-            if self.frees:
-                assert "__class__" in self.frees
-            return sorted(self.frees)
+            # If we're not nested we can't possibly have any free variables,
+            # as we can't close over class variables.  The exception to this
+            # rule is __class__, which we indeed can close over.
+            if '__class__' in self.frees:
+                return ['__class__']
+            return []
+
         free = {}
         free.update(self.frees)
         for name in self.uses.keys():
@@ -144,11 +158,17 @@ class Scope:
         for child in self.children:
             if child.name in self.explicit_globals:
                 child.global_scope = True
+            if child.nested:
+                frees = child.get_free_vars()
 
-            frees = child.get_free_vars()
-            globals = self.add_frees(frees)
-            for name in globals:
-                child.force_global(name)
+                globals = self.add_frees(frees)
+                for name in globals:
+                    child.force_global(name)
+            elif '__class__' in child.frees:
+                self.add_frees(['__class__'])
+            elif '__class__' in child.uses and '__class__' not in child.defs:
+                child.frees = {'__class__'}
+                self.add_frees(['__class__'])
 
     def force_global(self, name):
         """Force name to be global in scope.
@@ -181,7 +201,7 @@ class Scope:
         for name in names:
             sc = self.check_name(name)
             if self.nested:
-                if name == "__class__" and isinstance(self, ClassScope):
+                if name == "__class__":
                     self.cells[name] = 1
                 elif sc == SC_UNKNOWN or sc == SC_FREE \
                    or isinstance(self, ClassScope):
@@ -196,7 +216,11 @@ class Scope:
                 if name == "__class__":
                     if isinstance(self, ClassScope):
                         self.cells[name] = 1
-                elif sc == SC_LOCAL:
+                        continue
+                    elif self.findParentClass() is not None:
+                        self.frees[name] = 1
+                        continue
+                if sc == SC_LOCAL:
                     self.cells[name] = 1
                 elif sc != SC_CELL:
                     child_globals.append(name)
@@ -205,6 +229,15 @@ class Scope:
     def get_cell_vars(self):
         return sorted(self.cells.keys())
 
+    def findParentClass(self):
+        parent = self.parent
+        while not isinstance(parent, ClassScope):
+            if parent is None:
+                break
+            parent = parent.parent
+        return parent
+
+
 class ModuleScope(Scope):
     __super_init = Scope.__init__
 
@@ -246,17 +279,26 @@ class ClassScope(Scope):
     def __init__(self, name, module, lineno=0):
         self.__super_init(name, module, name, lineno=lineno)
 
-class SymbolVisitor:
+class SymbolVisitor(ASTVisitor):
     def __init__(self):
+        super().__init__()
         self.scopes = {}
+        self.module = None
         self.klass = None
 
+    def dump_scopes(self):
+        self.module.dump_recursive()
+
     # node that define new scopes
 
     def visitModule(self, node):
         scope = self.module = self.scopes[node] = ModuleScope()
         self.visit(node.body, scope)
 
+    def visitInteractive(self, node):
+        scope = self.module = self.scopes[node] = ModuleScope()
+        self.visit(node.body, scope)
+
     visitExpression = visitModule
 
     def visitFunctionDef(self, node, parent):
@@ -264,6 +306,7 @@ class SymbolVisitor:
             self.visit(node.decorator_list, parent)
         parent.add_def(node.name)
         scope = FunctionScope(node.name, self.module, self.klass, lineno=node.lineno)
+        scope.coroutine = isinstance(node, ast.AsyncFunctionDef)
         scope.parent = parent
         if parent.nested or isinstance(parent, FunctionScope):
             scope.nested = 1
@@ -283,11 +326,18 @@ class SymbolVisitor:
         ast.SetComp: "<setcomp>",
     }
 
-    def visitGeneratorExp(self, node, parent, assign=0):
+    def visitAwait(self, node, scope):
+        scope.coroutine = True
+        self.visit(node.value, scope)
+
+    def visitGeneratorExp(self, node, parent):
         scope = GenExprScope(
             self.module, self.klass, name=self._scope_names[type(node)],
             lineno=node.lineno
         )
+        if isinstance(node, ast.GeneratorExp):
+            scope.generator = True
+
         if parent.nested or isinstance(parent, FunctionScope) \
                 or isinstance(parent, GenExprScope):
             scope.nested = 1
@@ -316,7 +366,10 @@ class SymbolVisitor:
     visitDictComp = visitGeneratorExp
 
     def visitcomprehension(self, node, scope, is_outmost):
-        self.visit(node.target, scope, 1)
+        if node.is_async:
+            scope.coroutine = True
+
+        self.visit(node.target, scope)
         if is_outmost:
             scope.add_use(".0")
         else:
@@ -331,7 +384,7 @@ class SymbolVisitor:
         self.visit(node.expr, scope)
 
     def visitGenExprFor(self, node, scope):
-        self.visit(node.assign, scope, 1)
+        self.visit(node.assign, scope)
         self.visit(node.iter, scope)
         for if_ in node.ifs:
             self.visit(if_, scope)
@@ -339,11 +392,7 @@ class SymbolVisitor:
     def visitGenExprIf(self, node, scope):
         self.visit(node.test, scope)
 
-    def visitLambda(self, node, parent, assign=0):
-        # Lambda is an expression, so it could appear in an expression
-        # context where assign is passed.  The transformer should catch
-        # any code that has a lambda on the left-hand side.
-        assert not assign
+    def visitLambda(self, node, parent):
         scope = LambdaScope(self.module, self.klass, lineno=node.lineno)
         scope.parent = parent
         if parent.nested or isinstance(parent, FunctionScope):
@@ -384,6 +433,11 @@ class SymbolVisitor:
         scope.handle_children()
 
     def visitClassDef(self, node, parent):
+        if node.decorator_list:
+            self.visit(node.decorator_list, parent)
+        for kw in node.keywords:
+            self.visit(kw.value, parent)
+
         parent.add_def(node.name)
         for n in node.bases:
             self.visit(n, parent)
@@ -407,12 +461,8 @@ class SymbolVisitor:
 
     # name can be a def or a use
 
-    # XXX a few calls and nodes expect a third "assign" arg that is
-    # true if the name is being used as an assignment.  only
-    # expressions contained within statements may have the assign arg.
-
-    def visitName(self, node, scope, assign=0):
-        if assign or isinstance(node.ctx, ast.Store):
+    def visitName(self, node, scope):
+        if isinstance(node.ctx, ast.Store):
             scope.add_def(node.id)
         elif isinstance(node.ctx, ast.Del):
             # We do something to var, so even if we "undefine" it, it's a def.
@@ -422,36 +472,19 @@ class SymbolVisitor:
         else:
             scope.add_use(node.id)
 
-            if node.id == "__class__" and isinstance(scope, ClassScope):
-                # The idea is that in class body, __class__ is still bound to
-                # the outer containing class (if any). Otherwise, this class'
-                # __class__ is bound in methods of this class, but not class
-                # body itself (__class__ is created as result of class body
-                # evaluations).
-                parent = scope.parent
-                while not isinstance(parent, ClassScope):
-                    parent.frees["__class__"] = 1
-                    parent = parent.parent
-                parent.cells["__class__"] = 1
-                scope.frees["__class__"] = 1
-
-            if node.id == "super" and isinstance(scope, FunctionScope) \
-               and isinstance(scope.parent, ClassScope):
+            if node.id == "super" and isinstance(scope, FunctionScope):
                 # If super() is used, and special cell var __class__ to class
                 # definition, and free var to the method. This is complicated
                 # by the fact that original Python2 implementation supports
                 # free->cell var relationship only if free var is defined in
                 # a scope marked as "nested", which normal method in a class
                 # isn't.
-                # TODO: Check that super() takes no args.
-                scope.parent.add_def("__class__")
                 scope.add_use("__class__")
-                scope.frees["__class__"] = 1
 
     # operations that bind new names
 
     def visitFor(self, node, scope):
-        self.visit(node.target, scope, 1)
+        self.visit(node.target, scope)
         self.visit(node.iter, scope)
         self.visit(node.body, scope)
         if node.orelse:
@@ -480,16 +513,9 @@ class SymbolVisitor:
     def visitNonlocal(self, node, scope):
         # TODO: Check that var exists in outer scope
         for name in node.names:
-            scope.add_free(name)
+            scope.frees[name] = 1
             scope.nonlocals[name] = 1
 
-            if name == "__class__":
-                if isinstance(scope, FunctionScope) and isinstance(scope.parent, ClassScope):
-                    # TODO: Reconcile with handling for super() in visitName().
-                    scope.parent.cells["__class__"] = 1
-                    scope.frees["__class__"] = 1
-
-
     def visitAssign(self, node, scope):
         """Propagate assignment flag down to child nodes.
 
@@ -504,36 +530,50 @@ class SymbolVisitor:
         the assign flag to their children.
         """
         for n in node.targets:
-            self.visit(n, scope, 1)
+            self.visit(n, scope)
         self.visit(node.value, scope)
 
-    def visitAssName(self, node, scope, assign=1):
+    def visitAnnAssign(self, node, scope):
+        target = node.target
+        if isinstance(target, ast.Name):
+            if target.id in scope.nonlocals or target.id in scope.explicit_globals:
+                is_nonlocal = target.id in scope.nonlocals
+                raise SyntaxError(f"annotated name '{target.id}' can't be {'nonlocal' if is_nonlocal else 'global'}")
+            if node.simple or node.value:
+                scope.add_def(target.id)
+        else:
+            self.visit(node.target, scope)
+        self.visit(node.annotation, scope)
+        if node.value:
+            self.visit(node.value, scope)
+
+    def visitAssName(self, node, scope):
         scope.add_def(node.name)
 
-    def visitAssAttr(self, node, scope, assign=0):
-        self.visit(node.expr, scope, 0)
+    def visitAssAttr(self, node, scope):
+        self.visit(node.expr, scope)
 
-    def visitSubscript(self, node, scope, assign=0):
-        self.visit(node.value, scope, 0)
-        self.visit(node.slice, scope, 0)
+    def visitSubscript(self, node, scope):
+        self.visit(node.value, scope)
+        self.visit(node.slice, scope)
 
-    def visitAttribute(self, node, scope, assign=0):
-        self.visit(node.value, scope, 0)
+    def visitAttribute(self, node, scope):
+        self.visit(node.value, scope)
 
-    def visitSlice(self, node, scope, assign=0):
+    def visitSlice(self, node, scope):
         if node.lower:
-            self.visit(node.lower, scope, 0)
+            self.visit(node.lower, scope)
         if node.upper:
-            self.visit(node.upper, scope, 0)
+            self.visit(node.upper, scope)
         if node.step:
-            self.visit(node.step, scope, 0)
+            self.visit(node.step, scope)
 
     def visitAugAssign(self, node, scope):
         # If the LHS is a name, then this counts as assignment.
         # Otherwise, it's just use.
         self.visit(node.target, scope)
         if isinstance(node.target, ast.Name):
-            self.visit(node.target, scope, 1) # XXX worry about this
+            self.visit(node.target, scope)
         self.visit(node.value, scope)
 
     # prune if statements if tests are false
@@ -549,7 +589,12 @@ class SymbolVisitor:
     # a yield statement signals a generator
 
     def visitYield(self, node, scope):
-        scope.generator = 1
+        scope.generator = True
+        if node.value:
+            self.visit(node.value, scope)
+
+    def visitYieldFrom(self, node, scope):
+        scope.generator = True
         if node.value:
             self.visit(node.value, scope)
 
